/*****************************************************************************
*
* @brief Base class for a module of CloudBEAR SoC and auto-registration
*  and auto-creation infrastructure (factories).
*
* @author andrey.efimov@cloudbear.ru
*
* Copyright (c) 2016, 2020 CloudBEAR LLC, all rights reserved.
*
* This file contains confidential, proprietary information and trade
* secrets of CloudBEAR LLC. The information contained in this file
* may only be used by a person authorised under and to the extent
* permitted by a subsisting license agreement or design service
* agreement from CloudBEAR LLC.
*
* This entire notice must be reproduced on all copies of this file
* and copies of this file may only be made by a person if such person
* is permitted to do so under the terms of a subsisting license
* agreement or design service agreement from CloudBEAR LLC.
*
*****************************************************************************/

#ifndef MODULE_H_
#define MODULE_H_

#include <xcfg/ConfigNode.h>
#include <regexp_logger/logger_unit.h>
#include "ic_wire_interface.h"
#include "reg.h"
#include <map>
#include <cassert>

// Note [aefimov]: commented, because unused, but causes
// error: expected unqualified-id before ')' token in elfio.hpp:577
// #ifdef RISCV_ENABLE_ASSERT_AT_FAIL
//     #define fail() assert(0)
// #else
//     #define fail()
// #endif

// forward decl
class module_t;
class soc_t;

// Base module factory class
class module_factory_t {
public:
    virtual module_t* create(soc_t* parent, const std::string& name) = 0;
};

// Base module class
class module_t: public logger_unit_t {
public:
    module_t(soc_t* soc, const std::string& name);
    virtual ~module_t() { }

    // Non-copyable
    module_t(const module_t&) = delete;
    module_t& operator = (const module_t) = delete;

    // Module name
    const std::string& get_name() const { return name; }

    // Module type
    void set_type(const std::string& value) { type = value; }
    const std::string& get_type() const { return type; }

    // Initialization
    virtual bool init(ConfigNode /*config*/) { return true; }

    // Reset module
    virtual void reset() {};

    // Connect module to interrupt controller wire interface.
    virtual void connect_to_interrupt_controller(ic_wire_interface_t_p /*m*/) {
        // default implementation
    }

    // Slave port to SoC.
    // This port is used by SoC to forward load/store requests from other masters.
    // By default, does nothing.
    virtual bool load_callback(reg_t /*addr*/, size_t /*len*/, reg_t /*prv*/, uint8_t* /*bytes*/) {
        return false; }
    virtual bool store_callback(reg_t /*addr*/, size_t /*len*/, reg_t /*prv*/, const uint8_t* /*bytes*/) {
        return false; }
    virtual uint8_t* access_callback(reg_t /*addr*/, size_t /*align*/) { assert(0); return nullptr; }

    // Enable module
    virtual void start() { }

    // Disable module
    virtual void stop() { }

    // Perform one cycle of work
    virtual void tick(size_t /*cycle*/) { }

    // Perform one synchronization cycle
    // Called by SoC once in every $sync_interval$ ticks
    virtual void sync_tick(size_t /*cycle*/) { }

protected:
    soc_t* soc; // SoC instance
    std::string name; // Module name
    std::string type; // Module type

    // Device auto-registration and object creation stuff
public:
    static void register_type(const std::string& type, module_factory_t *factory) {
        // std::cout << "Registering '" << type << "' factory\n";
        get_factory()[type] = factory;
    }
    static module_t* create(const std::string& type, soc_t* soc, const std::string& name) {
        // std::cout << "Creating module of type '" << type << "' with id '" << name << "'\n";
        return get_factory().at(type)->create(soc, name);
    }
private:
    static std::map<std::string, module_factory_t*>& get_factory()
    {
        // Map should be created at first usage due to avoid races with
        // different construction order of global variables.
        // See https://stackoverflow.com/questions/13353519/access-violation-when-inserting-element-into-global-map
        static std::map<std::string, module_factory_t*> factories;
        return factories;
    }
};

// Specific factory creation.
// Invoke this once for each module type in module's cpp file.
#define REGISTER_MODULE_TYPE(name, cls) \
    class cls##_factory_t : public module_factory_t { \
    public: \
        cls##_factory_t() { \
            module_t::register_type(name, this); \
        } \
        virtual module_t *create(soc_t* soc, const std::string& mod_name) { \
            module_t* module = new cls(soc, mod_name); \
            module->set_type(name); \
            return module; \
        } \
    }; \
    static cls##_factory_t global_##cls##_factory_

#endif // MODULE_H_
